# Browser ready

module.exports = utils = {}

# Returns variable type as string
utils.typeOf = typeOf = (val) ->
	return switch Object::toString.call val
		when '[object Undefined]'	then 'undefined'
		when '[object String]'		then 'string'
		when '[object Function]'	then 'function'
		when '[object Boolean]'		then 'boolean'
		when '[object Object]'		then 'object'
		when '[object Array]'		then 'array'
		when '[object Null]'		then 'null'
		when '[object Number]'		then ( if val then 'number' else 'nan' )
		when '[object Date]'		then 'date'
		when '[object RegExp]'		then 'regexp'
		else 'null'

utils.sort = (array, fields...) ->
	array = array.slice()
	
	for field in fields
		field.order = !! ( field.order is 'asc' )
		field.key or field.key = ''
			
		do (field) -> array.sort (a, b) ->
			lesserReturn = if field.order then -1 else 1
			greaterReturn = if field.order then 1 else -1
			
			val1 = a[field.key] or ''
			val2 = b[field.key] or ''
			
			return lesserReturn if val1 < val2
			return greaterReturn if val1 > val2

			return 0

	return array

# Checks for object or array type
utils.isIterable = (vari, t) ->
	t = t or utils.typeOf vari

	return t is 'object' or t is 'array'

utils.getLength = (vari) ->
	t = utils.typeOf vari

	if t is 'array' or t is 'string'
		return vari.length
	else if t is 'object'
		length = 0
		++length for item of vari
		return length
	else
		return 0

utils.isNumber = (n) ->
	return not isNaN( parseFloat n ) and isFinite n

###
	Escapes all regexp characters in a string.
	
	@param str {String}
	@return {String} str
###
utils.escapeRegExp = (str) -> str.replace /[\-\[\]\/\{\}\(\)\*\+\?\.\\\^\$\|]/g, "\\$&"

###
	Converts a number to a boolean, on whether its an Integer of >= 1 or not

	@param number {Number}
	@return {Boolean}
###
utils.numberToBoolean = (number) -> if ( parseInt number ) >= 1 then true else false

###
	Converts a number to a boolean, on whether its an Integer of >= 1 or not

	@param number {Number}
	@return {Boolean}
###
utils.booleanToNumber = (boolean) -> if true then 1 else 0

###
	Format a string to be a url-friendly slug

	@param str {String} This "is" some text!
	@return {String} str this-is-some-text
###
utils.slugify = (str = '') ->
	str = decodeURIComponent str.toString('utf8')

	return str
		.toLowerCase()
		.replace( /[^a-z0-9-]+/g, '-' )
		.replace( /^[\s\-]+|[\s\-]+$/g, '' )
###
	Format a space delimited string to camelcase

	@param str {String} some words here
	@return {String} str someWordsHere
###
utils.camelCase = (str = '') ->
	return ( utils.slugify str )
		.replace /([ -]+)([a-zA-Z0-9])/g, (a, b, c) -> c.toUpperCase()
		.replace /([0-9]+)([a-zA-Z])/g, (a, b, c) -> c.toUpperCase()
		.replace /([0-9]+)([a-zA-Z])/g, (a, b, c) -> b + c.toUpperCase()

# Compares two file extensions
utils.isExt = (fileDir, ext) ->
	return false if not fileDir or typeof fileDir isnt 'string' or not ext

	regex = if ext?
		ext = ext.replace /^\./, ''
		new RegExp ".#{ext}$", 'i'
	else
		/\.([^\.]+)$/

	return fileDir?.match regex

utils.isUrl = (str = '') ->
	return str.match ///^
		(
			(ht|f)tps?\://
		|	www\.
		)
		[a-zA-Z0-9\-\.]+\.[a-zA-Z]{2,3}(/\S*)?
	///

# Dertermines if a string matches as an absolute directory
utils.isAbsolute		=
utils.isAbsolutePath	= (fileDir) ->
	return false if not fileDir
	return /^(\/|\w:)/.test fileDir

# Determines if a variable evalutes to false or empty
# TODO: improve this so that it works on array n does some basic stuff
# could optimize by not using isiterable and instead getting typeOf once
utils.empty = (vari, depth = 5) ->
	return true if depth < 1

	if utils.isIterable vari
		for key, val of vari
			if utils.isIterable val
				return false if not arguments.callee val, depth - 1
			else
				return false if val

		return true

	return ! vari

# Change an extension of a file path
utils.changeExt = (dir, ext = '') ->
	return dir if not dir

	ext = ext.replace /^\./, ''
	ext = '.' + ext if ext

	return dir.replace /\.[^\.]+$|$/, ext

# Minify a string, such as html
utils.minify = (str = '') ->
	return str
		.replace(/^\s+|\s+$/gm, '')
		.replace(/\n/g, '')
		.replace(/\s{2,}/g, ' ')

# Minify CSS
utils.minifyCss = (str = '') ->
	return str
		.replace(/^\s+|\s+$/gm, '')
		.replace(///
				\n
			|	\s+(?=\{)
			|	\;\s*(?=\})
		///g, '')
		.replace(/\s{2,}/g, ' ')
		.replace(/\s*\:\s+/g, ':')
		.replace(/\{\s+/g, '{')
		.replace(/\s+\}/g, '}')



# Merge obj1 by replacing values in obj1 with those from obj2. Iterates nested objects.
utils.merge	=
merge		= (obj1, obj2, depth = 8, iterators) ->
	iterators = iterators or [ 'object' ]

	return merge.merge obj1, obj2, depth, iterators

merge.merge = (obj1, obj2, depth, iterators) ->
	if depth > 0
		for own key of obj2
			obj2Type = utils.typeOf obj2[key]
			obj1Type = utils.typeOf obj1[key]

			if (
				( obj2Type in iterators ) and
				key of obj1 and
				( obj1Type in iterators )
			)
				arguments.callee obj1[key], obj2[key], depth - 1, iterators
			else
				obj1[key] = obj2[key]

	return obj1

# Merge obj1 and obj2 by replacing values where both objects share the same keys, preffering obj1
merge.by = (obj1, obj2, depth = 8, iterators, whiteList = true) ->
	iterators = iterators or [ 'object' ]
	return merge.by.merge obj1, obj2, depth, iterators, whiteList

merge.by.merge = (obj1, obj2, depth, iterators, whiteList) ->
	if depth > 0
		for own key of ( if not whiteList then obj2 else obj1 )
			continue if ( not whiteList and key of obj1 ) or ( whiteList and key not of obj2 )

			obj2Type = utils.typeOf obj2[key]
			obj1Type = utils.typeOf obj1[key]

			if ( obj2Type in iterators ) and ( obj1Type in iterators )
				arguments.callee obj1[key], obj2[key], depth - 1, iterators, whiteList
			else
				obj1[key] = obj2[key]

	return obj1

utils.merge.white = (obj1, obj2, depth, iterators) -> merge.by obj1, obj2, depth, iterators, true
utils.merge.black = (obj1, obj2, depth, iterators) -> merge.by obj1, obj2, depth, iterators, false


utils.clone	=
clone		= (obj, depth = 8) ->
	try
		return utils.clone.json obj
	catch err
		return utils.clone.merge obj, depth

clone.json = (obj) ->
	newObj = JSON.parse JSON.stringify obj
	return merge newObj, utils.clone.findRegExp obj

clone.findRegExp = (obj) ->
	results = {}
	for key, val of obj
		switch typeOf val
			when 'regexp'
				results[ key ] = val
			when 'object'
				results[ key ] = arguments.callee val

	return results

clone.merge = (obj, depth = 10, iterators) ->
	return merge {}, obj, depth, iterators

clone.merge.hard = (obj, depth = 10) ->
	return clone.merge.merge {}, obj, depth

clone.merge.merge = (obj1, obj2, depth = 10) ->
	if --depth > 0
		for key of obj2
			obj2Type = utils.typeOf obj2[key]

			wrap = null

			switch obj2Type
				when 'object'	then wrap = {}
				when 'array'	then wrap = []

			if wrap
				obj1[ key ] = clone.merge.merge wrap, obj2[ key ], depth
			else
				obj1[ key ] = obj2[ key ]

	return obj1

utils.clone.array = (obj) ->
	return obj.slice(0)

utils.bind = (obj, binding = obj, depth = 6) ->
	if depth > 0
		for key, val of obj
			switch utils.typeOf val
				when 'function'
					val		= utils.bind val, binding, depth - 1
					newFn	= val.bind binding

					utils.merge newFn, val
					obj[key] = newFn
				when 'object'
					utils.bind val, binding, depth - 1
	
	return obj

# Convert an object to an array. Especially useful when considering the "arguments" pseudo-array variable
utils.toArray = (args, sort = false) ->
	ary = Array::slice.call args
	
	if sort
		if utils.typeOf( sort ) is 'function'
			ary = ary.sort sort
		else
			ary = ary.sort()
			
	return ary

fullTypeTable = {
	'Undefined'	: undefined
	'Boolean'	: Boolean
	'String'	: String
	'Function'	: Function
	'Array'		: Array
	'Object'	: Object
	'Null'		: null
	'Number'	: Number
	'Date'		: Date
	'RegExp'	: RegExp
	'NaN'		: NaN
}

for key, val of fullTypeTable then do (key) ->
	key2	= key
	key		= key.toLowerCase()

	utils.typeOf[key]	=
	utils.typeOf[key2]	= (vari) -> return utils.typeOf( vari ) is key

# beta
utils.debounce = (o, fn) ->
	{ wait, maxWait, immediate } = o

	# TODO: make immediate work properly
	# it currently goes off twice when accruement goes off
	# may need to set a "firstTime" variable or something

	wait	= wait or 0
	maxWait	= maxWait or Infinity
	accrued	= 0
	timer	= hitTime = args = that = lastHitTime = undefined

	execute	= ->
		result = fn.apply that, args

		accrued = lastHitTime = 0
		reset()

		return result

	reset = -> clearTimeout timer

	check = ->
		hitTime = new Date()

		if lastHitTime
			accrued += hitTime - lastHitTime
		else if immediate
			lastHitTime	= hitTime
			return execute()

		lastHitTime	= hitTime

		if accrued >= maxWait
			return execute()
		else
			reset()
			timer = setTimeout execute, wait

		return null

	return ->
		args	= arguments
		that	= this
		return check()

